/*
 *  ANIMAL GUESSING GAME
 *
 *  This program prompt the user to think of an animal and then tries to 
 *  guess what it is by asking a series of questions. Really it's just an
 *  exercise in structures, pointers and tree balancing, and in packaging 
 *  and release structuring.
 *
 *  These is a reasonable diagrammatic explanation of what it does below.
 */

/*
 *  Very urgent things first.
 */

#define MAIN				/* headers to define RCS symbols for 
                                           inclusion in *this* object only */
					/* this file's own RCS symbols */
static char *animal_c_rcs_id = "$Header: /diskb/home/alexis/dev/animal/bin/RCS/animal.c,v 1.16 1999/01/07 11:37:44 alexis Exp $";

/*
 *  System-wide header files to include - with explanations why.
 */

#include <stdio.h>			/* for NULL, stderr, printf(), ... */
#include <malloc.h>			/* for calloc() */
#include <string.h>			/* strdup(), strlen(), ...  */
#include <stdlib.h>			/* exit() */
#include <errno.h>			/* for errno */
#include <signal.h>			/* for signal() and syms */

/*
 *  Program-wide header files to include - with explanations why.
 */

#include "utils.h"			/* for mybasename(), error(), ... */
#include "patchlevel.h"			/* for version info */

/*
 *  Some urgent symbols - these are forward declarations and structure defs.
 */

#define SUPPORT_REPHRASING	FALSE	/* support rephrasing? see note below */
#define SUPPORT_SLOVAK		FALSE	/* support slovak messages? */

/*
 *  This program uses a simple structure to contain information about 
 *  questions to ask or animals to guess, as well as information as to 
 *  what question to ask next or what animal to guess next depending on
 *  the answer to questions. The data structure can look this for example:
 *
 *                                Start Here
 *                             Has it got 4 legs?
 *                                    /\
 *                             (yes) /  \ (no)
 *                                  /    \
 *                                 /      \
 *                   It's a snake!?        Does it have long ears?
 *                                                    /\
 *                                             (yes) /  \ (no)
 *                                                  /    \
 *                                                 /      \
 *                                  It's a rabbit!?        It's a cow!?
 *
 *
 *  As you can see, the data structure 'hangs' by the first question, and 
 *  it can rapidly become quite unbalanced (e.g. if the user things of a 
 *  snake then one question is enough to ascertain it's a snake, but if the
 *  user were to think of a horse (how would the data structure be changed to
 *  incorporate a horse?) then the number of questions would be at least
 *  three. 
 *
 *  Now, rephrasing involves jiggling the questions about in order to
 *  turn this
 *
 *      /\
 *     * /\
 *      * /\
 *       *  *
 *
 *  into this
 *       _
 *      / \
 *     /\ /\
 *    * * * *
 *
 *  This 'rebalancing' is quite tricky, primarily because the questions
 *  themselves don't only get rearranged, but must also be changed! and the
 *  way they get changed depends precisely how the tree is inbalanced (there
 *  are four different sorts of inbalance including the one drawn above).
 *
 *  Rebalancing also requires that we keep a record of how out of balance 
 *  all sections of the tree are. This information is kept in the tree itself.
 *  Outofbalancedness is defined as the height of the tree hanging from the
 *  left branch minus the height of the tree hanging from the right branch.
 *  If the addition of new animals and questions causes an outofbalancedness
 *  to reach 2, then rebalancing is possible.
 * 
 *  However, the rephrasing results in questions that even the likes of
 *  Mr Spock might find confusing, so at the flick of a switch (namely
 *  the 'SUPPORT_REPHRASING' switch) support for rephrasing is removed.
 */

/*
 *  This program can also support the Slovak language. This requires
 *  words such as 'who' or 'what' to be used, but these words depend on
 *  the article type in the data structures (people require 'who', animals
 *  require 'what', places require 'where' etc). So the language of the
 *  data defines the language of the questions. (It'd be a bit daft if the
 *  program asked "Think of a zviera. Is it a konik?") So any datafiles that
 *  get initially loaded into the datastructures must tell the program not
 *  only what language to ask 'Think of a' in, but must also tell the
 *  program how to say 'animal' and 'who' or 'place' and where' in the 
 *  language of choice.
 *
 *  But, again, to simplify the program no end, (or at least simplify
 *  it for the casual reader,) support for foreign languages can be
 *  removed at the flick of the 'SUPPORT_SLOVAK' switch.
 *
 *  Note that comments are only in the non-SUPPORT_SLOVAK code.
 */

struct node {				/* information cell */
    char *txt;				/* for question text or animal name */
    struct node *lch;			/* where to go if answer is yes */
    struct node *rch;			/* where to go if answer is no */
#if SUPPORT_REPHRASING
    int hgt;				/* outofbalanceness for current node */
#endif
};

#define AN_RC_QUIT	   	1	/* return code used to mean quit */
#define AN_RC_RESTART      	2	/* to mean take it from the top */
#if SUPPORT_REPHRASING
#define AN_RC_RECALC       	3	/* to mean the structure has been added
					   to, and outofbalancedness *could*
					   now require rebalancing, but it 
					   might not, so recalculate to see */
#define AN_LGC_OR          	1       /* used in rephrasing questions */
#define AN_LGC_AND         	2	/* used in rephrasing questions */
#endif

#define AN_CH_LFT        	1	/* to tell rephraser where problem is */
#define AN_CH_RGT    		2	/* to tell rephraser where problem is */

/*
 *  Forward declarations for functions used only in this file.
 */

#if SUPPORT_REPHRASING			/* see loadfile() below for why return
					   type changes if rephrasing allowed */
static int  loadfile(char *, struct node **);
#else
static void loadfile(char *, struct node **);
#endif
static int  savefile(char *, struct node *);
static void usage(void);
static int  process(struct node **);
static int gen_lockfile_name(char *, char *);   /* write filename into buffer */

/*
 *  Data only used by functions in this file.
 */

static char *definer = NULL;		/* 'what', 'who' in right lang */
static char *class = NULL;		/* 'animal', 'person' in right lang */
#if SUPPORT_SLOVAK
static int language;			/* dialling code of lang's country */
#endif

/*
 *  Data belonging to this file, but referenced by others.
 */

char *progname;				/* name of this program */

/*
 *  Functions start here ...
 */

/******************************************************************************
*
*       Function:    main
*
*       Parameters:  command line parameter count
*                    array of command line words
*
*       Returns:     doesn't; it uses exit() instead
*
*       Description:
*
*                    initialises a few variables used elsewhere, processes
*                    command line options and required command line parameter
*	             - which should be a file name - by loading the specified
*                    file into the tree-like datastructure, then enters
*                    main loop in which it asks the user to think of an
*                    animal - or whatever - and then processes their response
*                    until they respond with 'q'.
*
*                    The way to follow what the program is doing is to
*                    draw the data structure with a pen and paper and then
*                    put your finger on the top node and move down asking
*                    questions until you get to a leaf-node when you state
*                    the guess therein. To determine the initial shape of the
*                    data structure then just play the game always saying
*                    "yes that's what it was" when it guesses.
*
******************************************************************************/

main(
int argc,
char *argv[])
{
    struct node *rootptr;		/* root if the data structure */
    int rc;				/* to temporily hold return codes */
    int lockingpid;			/* pid of current animal file locker */

					/* get name of this prog */
    progname = mybasename(argv[0], NULL);	
					/* enable error and warning messages */
    debug_level = (unsigned) DBG_ERROR|DBG_WARNING;

    /*
     *  Process valid command line options. Valid options are:
     *  -V		report version number and exit immediately
     *  -d <number>     set debugging level to specified value
     *  -v              enable info messages
     */

    while (argc > 1 && argv[1][0] == '-') {
        if (strcmp(argv[1], "-V") == 0) {
#if SUPPORT_SLOVAK
            if (language == 421)
                printf("%s verzia %s\n", progname, patchlevel);
            else
                printf("%s version %s\n", progname, patchlevel);
#else
            printf("%s version %s\n", progname, patchlevel);
#endif
            exit(EXIT_SUCCESS);

        } else if (strcmp(argv[1], "-v") == 0)
            debug_level = (unsigned) 3;

        else if (strcmp(argv[1], "-d") == 0) {
            if (argc<3 || sscanf(argv[2], "%lu", &debug_level)!=1) 
                usage();
            argc--; argv++;

        } else
            usage();

        argc--; 
        argv++;
    }

    /*
     *  Check there is one argument (name of data file) (in addition to
     *  the name of the program itself).
     */

    if (argc != 2) 
        usage();

    signal(SIGTERM, genericsighandler);
    signal(SIGINT,  genericsighandler);

    if ((lockingpid=my_lock(argv[1], gen_lockfile_name)) > 0)
        error("%s is in use by process %d", argv[1], lockingpid);
    else if (lockingpid < 0)
        error("can't create lockfile: %s", strerror(errno));

    rootptr = NULL;			/* mark tree empty */
    (void) loadfile(argv[1], &rootptr); /* populate tree with datafile */
    if (class == NULL) 
	error("class undefined (use 'C' in data file)");
    if (definer == NULL)
	error("definer undefined (use 'D' in data file)");


    /*
     *  Main loop - prompt user, process response, check return code
     *  for quit value.
     */

    for (rc=AN_RC_RESTART;rc!=AN_RC_QUIT;rc=process(&rootptr))
#if SUPPORT_SLOVAK
        if (language == 421)
            printf("Predstav si nejake %s ... ", class);
        else
            /* getart() generates the word 'a' or 'an' */
            printf("Think of %s%s ... ", getart(argv[1]), class);
#else
        /* getart() generates the word 'a' or 'an' */
        printf("Think of %s%s ... ", getart(argv[1]), class);
#endif

    /*
     *  Just before quitting save all the tree (which has probably changed).
     */

    savefile(argv[1], rootptr);
    exit(EXIT_SUCCESS);
}

/******************************************************************************
*
*       Function:    usage
*
*       Parameters:  none
*
*       Returns:     doesn't; it uses exit() instead
*
*       Description:
*
*                    this gets called when the user calls this program with
*                    invalid options of missing parameters. it tells them
*                    what they should have done and then exits.
*
******************************************************************************/

static void usage()
{
#if SUPPORT_SLOVAK
    if (language == 421) { 
        fprintf(stderr, "Pust mi tak: %s <subor.dat>\n", progname);
        fprintf(stderr, "             %s -V\n", progname);
    } else {
        fprintf(stderr, "Usage: %s [ -d <level> | -v ] <datafile>\n", progname);
        fprintf(stderr, "       %s -V\n", progname);
    }
#else
    fprintf(stderr, "Usage: %s [ -d <level> | -v ] <datafile>\n", progname);
    fprintf(stderr, "       %s -V\n", progname);
#endif
    exit(EXIT_FAILURE);
}

/******************************************************************************
*
*       Function:    rotate
*
*       Parameters:  three addresses of node pointers
*
*       Returns:     nothing
*
*       Description:
*
*                    swaps three values around - this is used by the
*                    rephrasing code, and isn't compiled in if rephrasing
*                    is not included.
*
******************************************************************************/

#if SUPPORT_REPHRASE
static void rotate(
struct node **ptr1_addr,
struct node **ptr2_addr,
struct node **ptr3_addr)
{
    struct node *ptr;

    ptr = *ptr1_addr;
    *ptr1_addr = *ptr2_addr;
    *ptr2_addr = *ptr3_addr;
    *ptr3_addr = ptr;
}
#endif

/******************************************************************************
*
*       Function:    process
*
*       Parameters:  address of a node to start at
*                    (note function is recursive (i.e. calls itself) so
*                    while *initially* it's passed the very top node, 
*                    later it's passed a lower node. Note it's passed
*                    *the address of* a node, this is order that it can
*                    change the node itself itself!
*
*       Returns:     code indicating what to do as it reverses up the data
*
*       Description:
*
*                   Ok, this is the biggy! What it's basically doing is
*                   asking a question and if the answer it receives leads it
*                   to ask another question then it calls itself to get itself
*                   to do that (rather than doing it itself)! Follow? :-)
*                   In the event that it encounters a non-question - i.e.
*                   an animal or a person or whatever, then it says "I think
*                   I know what you were thinking of!" and proceeds to 
*                   state its guess and ask if it was correct. If it was
*                   correct, well there's nothing else to do except retreat
*                   back out of the calls to itself it made on the way down.
*                   If it's wrong then it needs to ask what the user was 
*                   thinking of and replace the animal that it thought the
*                   user was thinking of with a question that differentiates
*                   between the guess the program made and what the user's
*                   animal actually was. Having inserted the question where
*                   the animal was, it then need to create two new nodes
*                   for the animals and insert them. 
*
*                   And if it did create two new nodes then there is a change
*                   that unbalancedness has been introduced, so we must
*                   request that calls of this function check for it as
*                   they retreat out of the recursion. And if they do find
*                   outofbalancedness of level 2 (1 is acceptable) then they
*                   have to do some rephrasing - if it's switched on.
*                   
******************************************************************************/

static int process(
struct node **nodeptr_addr)
{
    char response[128], response2[128];
    int rc;
#if SUPPORT_REPHRASING
    int rc, lhgt, rhgt, llhgt, lrhgt, rlhgt, rrhgt;
#endif

#if SUPPORT_REPHRASING
    debug(DBG_SPECIAL1, "height of current tree is %d", (*nodeptr_addr)->hgt);
#endif

    /*
     *  If the current node has children then it must be a question.
     */

    if ((*nodeptr_addr)->lch != NULL) {
#if SUPPORT_SLOVAK
        if (language == 421) {
            printf("%s? [anzk] ", (*nodeptr_addr)->txt);
            (void) fgets(response, (int) sizeof(response), stdin);
            switch (response[0]) {
                case 'a': rc = process(&(*nodeptr_addr)->lch);
                          break;
                case 'n': rc = process(&(*nodeptr_addr)->rch);
                          break;
                case 'k': return(AN_RC_QUIT);
                case 'z': return(AN_RC_RESTART);
            }
        } else {
            printf("%s? [ynqar] ", (*nodeptr_addr)->txt);
            (void) fgets(response, (int) sizeof(response), stdin);
            switch (response[0]) {
                case 'y': rc = process(&(*nodeptr_addr)->lch);
                          break;
                case 'n': rc = process(&(*nodeptr_addr)->rch);
                          break;
                case 'q': return(AN_RC_QUIT);
                case 'r': return(AN_RC_RESTART);
                case 'a': exit(EXIT_SUCCESS);
            }
        }
#else
        /*
         *  Ask the question and get the answer
         */

        printf("%s? [ynqar] ", (*nodeptr_addr)->txt);
        (void) fgets(response, (int) sizeof(response), stdin);

        /*
         *  If the user answers 'yes' to the question then move down to
         *  the left and call this function again. If 'no' the go to the
         *  right.
         */

        switch (response[0]) {
            case 'y': rc = process(&(*nodeptr_addr)->lch);
                      break;
            case 'n': rc = process(&(*nodeptr_addr)->rch);
                      break;
            case 'q': return(AN_RC_QUIT);
            case 'r': return(AN_RC_RESTART);
            case 'a': exit(EXIT_SUCCESS);
        }
#endif

        /*
         *  Now, the call to this function itself above resulted (after
         *  a few more deeper calls to itself) to the animal being correctly
         *  guessed or the guess being wrong and a new question and a new
         *  animal being inserted into the data structure. Note that we
         *  saved the record code. Now if the return code says data was added
         *  and hence we need to recalculate outofbalancednesses then that's
         *  what we do. Outofbalancednesses or 2 or -2 require rebalancing.
         *  1 or -1 is not unbalanced enoug to be able to rebalance it, and
         *  0 is perfectly balanced already.
         *
         *  The code for actually doing the rebalancing is a nightmare!
         *  And explaining how rebalancing is done is just impossible; but
         *  basically it involves rearranging three questions and putting
         *  'or's and 'not's and 'and's in front of some of them.
         */

        debug(DBG_SPECIAL1, "process: coming up from '%c' branch, rc is %d", response[0], rc);
        switch (rc) {
#if SUPPORT_REPHRASING
            case AN_RC_RECALC:  
                debug(DBG_SPECIAL1, "process: reblnc ...");
                lhgt = (*nodeptr_addr)->lch->hgt;
                rhgt = (*nodeptr_addr)->rch->hgt;
                (*nodeptr_addr)->hgt = ((lhgt > rhgt) ? lhgt : rhgt) + 1;
                switch (rhgt - lhgt) {
                    case -2:
                        debug(DBG_SPECIAL1, "process: reblnc -2");
                        llhgt = (*nodeptr_addr)->lch->lch->hgt;
                        lrhgt = (*nodeptr_addr)->lch->rch->hgt;
                        switch (lrhgt - llhgt) {
                            case -1:
                                debug(DBG_SPECIAL1, "process: reblnc -2/-1");
                                /* rotate(&(*nodeptr_addr)->lch->rch,
                                       nodeptr_addr,
                                       &(*nodeptr_addr)->lch);
                                (void) rephrase(TRUE, AN_LGC_AND, FALSE, nodeptr_addr, AN_CH_LFT); */
                                rotate(&(*nodeptr_addr)->lch->rch,
                                       nodeptr_addr,
                                       &(*nodeptr_addr)->lch);
                                (void) rephrase(TRUE, AN_LGC_AND, TRUE, nodeptr_addr, AN_CH_RGT);
                                return(AN_RC_RESTART);
                            case 0:
                                debug(DBG_SPECIAL1, "process: reblnc -2/0");
                                error("not coded yet");
                            case 1:
                                debug(DBG_SPECIAL1, "process: reblnc -2/1");
                                rotate(&(*nodeptr_addr)->lch->lch,
                                       nodeptr_addr,
                                       &(*nodeptr_addr)->lch);
                                (void) rephrase(TRUE, AN_LGC_OR, FALSE, nodeptr_addr, AN_CH_LFT);
                                return(AN_RC_RESTART);
                            default:
                                error("process: reblnc -2/default");
                        }
                    case -1:
                        debug(DBG_SPECIAL1, "process: reblnc -1");
                        return(AN_RC_RECALC);
                    case 0:
                        debug(DBG_SPECIAL1, "process: reblnc 0");
                        return(AN_RC_RECALC);
                    case 1:
                        debug(DBG_SPECIAL1, "process: reblnc 1");
                        return(AN_RC_RECALC);
                    case 2:
                        debug(DBG_SPECIAL1, "process: reblnc 2");
                        rlhgt = (*nodeptr_addr)->rch->lch->hgt;
                        rrhgt = (*nodeptr_addr)->rch->rch->hgt;
                        switch(rrhgt - rlhgt) {
                            case -1: 
                                debug(DBG_SPECIAL1, "process: reblnc 2/-1");
                                rotate(&(*nodeptr_addr)->rch->rch,
                                       nodeptr_addr,
                                       &(*nodeptr_addr)->rch);
				(void) rephrase(TRUE, AN_LGC_AND, FALSE, nodeptr_addr, AN_CH_RGT);
                                return(AN_RC_RESTART);
                            case 0:
                                debug(DBG_SPECIAL1, "process: reblnc 2/0");
                                error("unexpected situation (001)");
                            case 1:
                                debug(DBG_SPECIAL1, "process: reblnc 2/1");
                                rotate(&(*nodeptr_addr)->rch->lch,
                                       nodeptr_addr,
                                       &(*nodeptr_addr)->rch);
				(void) rephrase(TRUE, AN_LGC_OR, TRUE, nodeptr_addr, AN_CH_LFT);
                                return(AN_RC_RESTART);
                            default:
                                error("unexpected situation (002)");
                        }
                    default:
                        error("unexpected situation (003)");
                }
                return(AN_RC_RECALC);
#endif
            case AN_RC_QUIT:    
                return(AN_RC_QUIT);
            case AN_RC_RESTART: 
                return(AN_RC_RESTART);
        }
    } else {
#if SUPPORT_SLOVAK
        if (language == 421) {
            printf("Je to %s? [ankz] ", (*nodeptr_addr)->txt);
            (void) fgets(response, (int) sizeof(response), stdin);
            switch (response[0]) {
                case 'a': printf("Mudry som nie?\n");
                          return(AN_RC_RESTART);
                case 'z': return(AN_RC_RESTART);
                case 'k': return(AN_RC_QUIT);
                case 'n': printf("%s je to? [kz] ", definer);
                          (void) fgets(response, (int) sizeof(response), stdin);
                          switch (response[0]) {
                              case 'k': return(AN_RC_QUIT);
                              case 'z': return(AN_RC_RESTART);
                          }
                          printf("Pis mi otazku; odpoved pre %s musi byt 'ano', a pre %s musi byt 'nie': [kz] ", (*nodeptr_addr)->txt, response);
                          (void) fgets(response2, (int) sizeof(response2), stdin);
                          switch (response2[0]) {
                              case 'k': return(AN_RC_QUIT);
                              case 'z': return(AN_RC_RESTART);
                          }
#ifdef MSDOS
                          /*
                           *  This program works under MSDOS, except that
                           *  MSDOS uses CR-NL to represent what Unix does
                           *  with only NL. So we need to erase an extra
                           *  character on the end of an user-supplied
                           *  question for MSDOS.
                           */

                          if (response2[strlen(response2)-1] == '\r') 
                              response2[strlen(response2)-1] = '\0';
                          if (response[strlen(response)-1] == '\r') 
                              response[strlen(response)-1] = '\0';
#endif
                          if (response[strlen(response)-1] == '\n') 
                              response[strlen(response)-1] = '\0';
                          if (response2[strlen(response2)-1] == '\n') 
                              response2[strlen(response2)-1] = '\0';
                          if (response2[strlen(response2)-1] == '?') 
                              response2[strlen(response2)-1] = '\0';
                          (*nodeptr_addr)->lch=calloc(1, sizeof(struct node));
                          (*nodeptr_addr)->lch->txt = (*nodeptr_addr)->txt;
                          (*nodeptr_addr)->lch->lch = NULL;
                          (*nodeptr_addr)->lch->rch = NULL;
#if SUPPORT_REPHRASING
                          (*nodeptr_addr)->lch->hgt = 0;
#endif
                          (*nodeptr_addr)->rch=calloc(1, sizeof(struct node));
                          (*nodeptr_addr)->rch->txt=strdup(response);
                          (*nodeptr_addr)->rch->lch = NULL;
                          (*nodeptr_addr)->rch->rch = NULL;
#if SUPPORT_REPHRASING
                          (*nodeptr_addr)->rch->hgt = 0;
#endif
                          (*nodeptr_addr)->txt = strdup(response2);
#if SUPPORT_REPHRASING
                          (*nodeptr_addr)->hgt = 1;
#endif
                          printf("Diky! Uz som bol mudry, ale teraz som mudrejsi!\n");
#if SUPPORT_REPHRASING
                          return(AN_RC_RECALC);
#else
                          return(AN_RC_RESTART);
#endif
            }
        } else {
            /* getart() generates the word 'a' or 'an' */
            printf("Is it %s%s? [ynqar] ", getart((*nodeptr_addr)->txt), (*nodeptr_addr)->txt);
            (void) fgets(response, (int) sizeof(response), stdin);
            switch (response[0]) {
                case 'y': printf("Great!\n");
                          return(AN_RC_RESTART);
                case 'r': return(AN_RC_RESTART);
                case 'q': return(AN_RC_QUIT);
                case 'a': exit(EXIT_SUCCESS);
                case 'n': printf("%s is it? [qar] ", definer);
                          (void) fgets(response, (int) sizeof(response), stdin);
                          switch (response[0]) {
                              case 'q': return(AN_RC_QUIT);
                              case 'a': exit(EXIT_SUCCESS);
                              case 'r': return(AN_RC_RESTART);
                          }
                          printf("Enter a question for which the answer for %s is yes, and for which %s is no: [qar] ", (*nodeptr_addr)->txt, response);
                          (void) fgets(response2, (int) sizeof(response2), stdin);
                          switch (response2[0]) {
                              case 'q': return(AN_RC_QUIT);
                              case 'a': exit(EXIT_SUCCESS);
                              case 'r': return(AN_RC_RESTART);
                          }
#ifdef MSDOS
                          /*
                           *  This program works under MSDOS, except that
                           *  MSDOS uses CR-NL to represent what Unix does
                           *  with only NL. So we need to erase an extra
                           *  character on the end of an user-supplied
                           *  question for MSDOS.
                           */

                          if (response2[strlen(response2)-1] == '\r') 
                              response2[strlen(response2)-1] = '\0';
                          if (response[strlen(response)-1] == '\r') 
                              response[strlen(response)-1] = '\0';
#endif
                          if (response[strlen(response)-1] == '\n') 
                              response[strlen(response)-1] = '\0';
                          if (response2[strlen(response2)-1] == '\n') 
                              response2[strlen(response2)-1] = '\0';
                          if (response2[strlen(response2)-1] == '?') 
                              response2[strlen(response2)-1] = '\0';
                          (*nodeptr_addr)->lch=calloc(1, sizeof(struct node));
                          (*nodeptr_addr)->lch->txt = (*nodeptr_addr)->txt;
                          (*nodeptr_addr)->lch->lch = NULL;
                          (*nodeptr_addr)->lch->rch = NULL;
#if SUPPORT_REPHRASING
                          (*nodeptr_addr)->lch->hgt = 0;
#endif
                          (*nodeptr_addr)->rch=calloc(1, sizeof(struct node));
                          (*nodeptr_addr)->rch->txt=strdup(response);
                          (*nodeptr_addr)->rch->lch = NULL;
                          (*nodeptr_addr)->rch->rch = NULL;
#if SUPPORT_REPHRASING
                          (*nodeptr_addr)->rch->hgt = 0;
#endif
                          (*nodeptr_addr)->txt = strdup(response2);
#if SUPPORT_REPHRASING
                          (*nodeptr_addr)->hgt = 1;
#endif
                          printf("Thanks! I was intelligent but now I'm more intelligent!\n");
#if SUPPORT_REPHRASING
                          return(AN_RC_RECALC);
#else
                          return(AN_RC_RESTART);
#endif
            }
        }
#else
        /* getart() generates the word 'a' or 'an' */
        printf("Is it %s%s? [ynqar] ", getart((*nodeptr_addr)->txt), (*nodeptr_addr)->txt);
        (void) fgets(response, (int) sizeof(response), stdin);
        switch (response[0]) {
            case 'y': printf("Great!\n");
                      return(AN_RC_RESTART);
            case 'r': return(AN_RC_RESTART);
            case 'q': return(AN_RC_QUIT);
            case 'a': exit(EXIT_SUCCESS);

            /*  
             *  Here we are asking 'what is it?'. We get the new animal
             *  and a question to differentiate between the wrong guess and
             *  what the user actually was thinking of.
             */

            case 'n': printf("%s is it? [qar] ", definer);
                      (void) fgets(response, (int) sizeof(response), stdin);
                      switch (response[0]) {
                          case 'q': return(AN_RC_QUIT);
                          case 'a': exit(EXIT_SUCCESS);
                          case 'r': return(AN_RC_RESTART);
                      }
                      printf("Enter a question for which the answer for %s is yes, and for which %s is no: [qar] ", (*nodeptr_addr)->txt, response);
                      (void) fgets(response2, (int) sizeof(response2), stdin);
                      switch (response2[0]) {
                          case 'q': return(AN_RC_QUIT);
                          case 'a': exit(EXIT_SUCCESS);
                          case 'r': return(AN_RC_RESTART);
                      }
#ifdef MSDOS
                      /*
                       *  This program works under MSDOS, except that
                       *  MSDOS uses CR-NL to represent what Unix does
                       *  with only NL. So we need to erase an extra
                       *  character on the end of an user-supplied
                       *  question for MSDOS.
                       */

                      if (response2[strlen(response2)-1] == '\r') 
                          response2[strlen(response2)-1] = '\0';
                      if (response[strlen(response)-1] == '\r') 
                          response[strlen(response)-1] = '\0';
#endif
                      if (response[strlen(response)-1] == '\n') 
                          response[strlen(response)-1] = '\0';
                      if (response2[strlen(response2)-1] == '\n') 
                          response2[strlen(response2)-1] = '\0';
                      if (response2[strlen(response2)-1] == '?') 
                          response2[strlen(response2)-1] = '\0';

                      /*
                       *  Here we create a new left child node, and copy
                       *  the wrongly guessed animal into it.
                       */

                      (*nodeptr_addr)->lch=calloc(1, sizeof(struct node));
                      (*nodeptr_addr)->lch->txt = (*nodeptr_addr)->txt;
                      (*nodeptr_addr)->lch->lch = NULL;
                      (*nodeptr_addr)->lch->rch = NULL;
#if SUPPORT_REPHRASING
                      (*nodeptr_addr)->lch->hgt = 0;
#endif
                      /*
                       *  Here we create a new right child node, and copy
                       *  the new animal into it.
                       */

                      (*nodeptr_addr)->rch=calloc(1, sizeof(struct node));
                      (*nodeptr_addr)->rch->txt=strdup(response);
                      (*nodeptr_addr)->rch->lch = NULL;
                      (*nodeptr_addr)->rch->rch = NULL;
#if SUPPORT_REPHRASING
                      (*nodeptr_addr)->rch->hgt = 0;
#endif
                      /*
                       *  And where the wrong guess was, we now put the
                       *  question that differentiates between the two 
                       *  animals.
                       */

                      (*nodeptr_addr)->txt = strdup(response2);
#if SUPPORT_REPHRASING
                      (*nodeptr_addr)->hgt = 1;
#endif
                      printf("Thanks! I was intelligent but now I'm more intelligent!\n");
#if SUPPORT_REPHRASING
                      /*
                       *  If rephrasing is enabled, then we need to tell
                       *  the process()es that called us, to check for 
                       *  imbalance.
                       */

                      return(AN_RC_RECALC);
#else
                      return(AN_RC_RESTART);
#endif
        }
#endif
    }
}

/******************************************************************************
*
*       Function:    rephrase
*
*       Parameters:  don't ask
*
*       Returns:     nothing worth noting
*
*       Description:
*
*                    oh my god ... you just don't want to know what this
*                    does ... I'm not sure I can even remember!
*
******************************************************************************/

#if SUPPORT_REPHRASING
static int rephrase(
int logic1,
int comb,
int logic2,
struct node **nodeptr_addr,
int link)
{
    char response[128], response2[128];
    struct node *tmpptr;

#if SUPPORT_SLOVAK
    if (language == 421) {
        printf("Aby moje otazky boli jasnejsie, mohol by si: \n");
        printf("    1) zjasnesit toto: '(%s%s) %s (%s%s)'\n", 
           (logic1 ? "" : "NOT "), 
           (*nodeptr_addr)->txt,
           ((comb==AN_LGC_AND) ? "A" : "ALEBO"), 
           (logic2 ? "" : "NIE "),
           ((link == AN_CH_RGT) ? (*nodeptr_addr)->rch : (*nodeptr_addr)->lch)->txt);
        printf("    2) or zjasnesit toto: '(%s%s) %s (%s%s)'\n",
           (logic1 ? "NIE " : ""),
           (*nodeptr_addr)->txt, 
           ((comb==AN_LGC_AND) ? "ALEBO" : "A"), 
           (logic2 ? "NIE " : ""),
           ((link == AN_CH_RGT) ? (*nodeptr_addr)->rch : (*nodeptr_addr)->lch)->txt);

        printf("    3) nerobit nic, ale budem pouzivat prvu moznost\n");
        printf("Vyber si [123]: ");
    } else {
        printf("In order to simplify the questions I ask, you may: \n");
        printf("    1) rephrase this: '(%s%s) %s (%s%s)'\n", 
           (logic1 ? "" : "NOT "), 
           (*nodeptr_addr)->txt,
           ((comb==AN_LGC_AND) ? "AND" : "OR"), 
           (logic2 ? "" : "NOT "),
           ((link == AN_CH_RGT) ? (*nodeptr_addr)->rch : (*nodeptr_addr)->lch)->txt);
        printf("    2) or rephrase this: '(%s%s) %s (%s%s)'\n",
           (logic1 ? "NOT " : ""),
           (*nodeptr_addr)->txt, 
           ((comb==AN_LGC_AND) ? "OR" : "AND"), 
           (logic2 ? "NOT " : ""),
           ((link == AN_CH_RGT) ? (*nodeptr_addr)->rch : (*nodeptr_addr)->lch)->txt);
        printf("    3) not rephrase either of them\n");
        printf("Choose [123]: ");
    }
#else
    printf("In order to simplify the questions I ask, you may: \n");
    printf("    1) rephrase this: '(%s%s) %s (%s%s)'\n", 
       (logic1 ? "" : "NOT "), 
       (*nodeptr_addr)->txt,
       ((comb==AN_LGC_AND) ? "AND" : "OR"), 
       (logic2 ? "" : "NOT "),
       ((link == AN_CH_RGT) ? (*nodeptr_addr)->rch : (*nodeptr_addr)->lch)->txt);
    printf("    2) or rephrase this: '(%s%s) %s (%s%s)'\n",
       (logic1 ? "NOT " : ""),
       (*nodeptr_addr)->txt, 
       ((comb==AN_LGC_AND) ? "OR" : "AND"), 
       (logic2 ? "NOT " : ""),
       ((link == AN_CH_RGT) ? (*nodeptr_addr)->rch : (*nodeptr_addr)->lch)->txt);
    printf("    3) not rephrase either of them\n");
    printf("Choose [123]: ");
#endif

    (void) fgets(response, (int) sizeof(response), stdin);
    if (response[0] == '\0' || response[0] == '3') {
#if SUPPORT_SLOVAK
        if (language == 421) {
            sprintf(response2, "(%s%s) %s (%s%s)'",
                (logic1 ? "" : "NIE "), 
                (*nodeptr_addr)->txt, 
                ((comb==AN_LGC_AND) ? "A" : "ALEBO"), 
                (logic2 ? "" : "NIE "),
                ((link == AN_CH_RGT) ? (*nodeptr_addr)->rch : (*nodeptr_addr)->lch)->txt);
        } else {
            sprintf(response2, "(%s%s) %s (%s%s)'",
               (logic1 ? "" : "NOT "), 
               (*nodeptr_addr)->txt, 
               ((comb==AN_LGC_AND) ? "AND" : "OR"), 
               (logic2 ? "" : "NOT "),
               ((link == AN_CH_RGT) ? (*nodeptr_addr)->rch : (*nodeptr_addr)->lch)->txt);
        }
#else
        sprintf(response2, "(%s%s) %s (%s%s)'",
           (logic1 ? "" : "NOT "), 
           (*nodeptr_addr)->txt, 
           ((comb==AN_LGC_AND) ? "AND" : "OR"), 
           (logic2 ? "" : "NOT "),
           ((link == AN_CH_RGT) ? (*nodeptr_addr)->rch : (*nodeptr_addr)->lch)->txt);
#endif
    } else if (response[0] == '1' || response[0] == '2') {
#if SUPPORT_SLOVAK
        if (language == 421)
            printf("Tak zjasnesaj to: ");
        else
            printf("Enter the rephrasing: ");
#else
        printf("Enter the rephrasing: ");
#endif
        (void) fgets(response2, (int) sizeof(response2), stdin);
        if (response[0] == '2') {
            debug(DBG_SPECIAL1, "swapping");
            tmpptr = (*nodeptr_addr)->lch;
            (*nodeptr_addr)->lch = (*nodeptr_addr)->rch;
            (*nodeptr_addr)->rch = tmpptr;
        } else
            debug(DBG_SPECIAL1, "not swapping");
    }
    free((*nodeptr_addr)->txt);
    (*nodeptr_addr)->txt = strdup(response2);
    return(0);
}
#endif

/******************************************************************************
*
*       Function:    savefile
*
*       Parameters:  name of file to save tree to
*                    pointer to top of data structure to save
*
*       Returns:     nothing worth noting
*
*       Description: 
*                    saves the words 'what' and 'animal' (or 'who' and
*                    'person' or whatever) and then saves the top question
*                    and then calls itself to save the whole left sub-tree
*                    and then calls itself again to save to whole right
*                    sub-tree. Of course if the current node is not a question
*                    then it just saves the animal (or whatever) and has 
*                    no sub-trees underneath to deal with.
*
*                    Note the only the top invocation actually opens the 
*                    file - all the others know it's open already - and they
*                    know it's open by examining the file handle - hence it
*                    must be either global or function-static.
*
******************************************************************************/

static int savefile(
char *datafile,
struct node *nodeptr)
{
    static FILE *fp = NULL;
    if (fp == NULL) {
        if ((fp=fopen(datafile, "w")) == NULL)
#if SUPPORT_SLOVAK
            if (language == 421)
                error("nemozem nahrat subor");
            else
                error("can't save datafile");
#else
            error("can't save datafile");
#endif
        fprintf(fp, "D %s\n", definer);
        fprintf(fp, "C %s\n", class);
#if SUPPORT_SLOVAK
        fprintf(fp, "L %d\n", language);
#endif
    }
    
    if (nodeptr->lch == NULL)
        fprintf(fp, "A %s\n", nodeptr->txt);
    else {
        fprintf(fp, "Q %s\n", nodeptr->txt);
        savefile(NULL, nodeptr->lch);
        savefile(NULL, nodeptr->rch);
    }
}

/******************************************************************************
*
*       Function:    loadfile
*
*       Parameters:  name of file to load tree from
*                    address of pointer to top of data structure to save
*                    (note it's the 'address of' since this function will
*                    want to change its value).
*
*       Returns:     nothing worth noting
*
*       Description: 
*                    loads the words 'what' and 'animal' (or 'who' and
*                    'person' or whatever) and then loads the top question
*                    and then calls itself to load the whole left sub-tree
*                    and then calls itself again to load to whole right
*                    sub-tree. Of course if the current node is not a question
*                    then it just loads the animal (or whatever) and has 
*                    no sub-trees underneath to deal with.
*
*                    Note the only the top invocation actually opens the 
*                    file - all the others know it's open already - and they
*                    know it's open by examining the file handle - hence it
*                    must be either global or function-static.
*
******************************************************************************/

#if SUPPORT_REPHRASING
static int loadfile(
#else
static void loadfile(
#endif
char *datafile,
struct node **nodeptr_addr)
{
    static FILE *fp = NULL;
    char text[128];
#if SUPPORT_REPHRASING
    int lhgt, rhgt;
#endif

    if (fp == NULL && (fp=fopen(datafile, "r")) == NULL)
#if SUPPORT_SLOVAK
        if (language == 421)
            error("nemohol som citat %s", datafile);
        else
            error("couldn't open %s", datafile);
#else
        error("couldn't open %s", datafile);
#endif

    while (!feof(fp)) {
        (void) fgets(text, (int) sizeof(text), fp);
        if (text[strlen(text)-1] == '\n')
            text[strlen(text)-1] = '\0';
        if (text[strlen(text)-1] == '\?')
            text[strlen(text)-1] = '\0';
        
        switch (text[0]) {
            case '\0': 
            case '#': break;
            case 'Q':
            case 'A':
                *nodeptr_addr = calloc(1, sizeof(struct node));
                (*nodeptr_addr)->txt = strdup(text+2);
                if (text[0] == 'Q') {
#if SUPPORT_REPHRASING
                    lhgt = loadfile(NULL, &(*nodeptr_addr)->lch);
                    rhgt = loadfile(NULL, &(*nodeptr_addr)->rch);
                    (*nodeptr_addr)->hgt = ((lhgt > rhgt) ? lhgt : rhgt) + 1;
#else
                    loadfile(NULL, &(*nodeptr_addr)->lch);
                    loadfile(NULL, &(*nodeptr_addr)->rch);
#endif
                } else {
                    (*nodeptr_addr)->lch = NULL;
                    (*nodeptr_addr)->rch = NULL;
#if SUPPORT_REPHRASING
                    (*nodeptr_addr)->hgt = 0;
#endif
                }
#if SUPPORT_REPHRASING
                return((*nodeptr_addr)->hgt);
#else
                return;
#endif
            case 'C':
                class = strdup(text+2);
                debug(DBG_SPECIAL1, "class set to %s", class);
                break;
#if SUPPORT_SLOVAK
            case 'L':
                language = atoi(text+2);
                debug(DBG_SPECIAL1, "language set to %d", language);
                break;
#endif
            case 'D':
                definer = strdup(text+2);
                debug(DBG_SPECIAL1, "definer set to %s", definer);
                break;
        }
    }

    if (feof(fp))
        fclose(fp);

#if SUPPORT_REPHRASING
    return((*nodeptr_addr)->hgt);
#endif
}

int gen_lockfile_name(
char *lockfile,
char *lockwhat)
{
    sprintf(lockfile, "%s.pid", lockwhat);
    return(0);
}
